from PyQt5.QtCore import QRegExp
from PyQt5.QtGui import QColor, QTextCharFormat, QFont, QSyntaxHighlighter


def format(color, style=''):
    """
    Return a QTextCharFormat with the given attributes.
    """
    _color = QColor()
    if type(color) is not str:
        _color.setRgb(color[0], color[1], color[2])
    else:
        _color.setNamedColor(color)

    _format = QTextCharFormat()
    _format.setForeground(_color)
    if 'bold' in style:
        _format.setFontWeight(QFont.Bold)
    if 'italic' in style:
        _format.setFontItalic(True)

    return _format


# Syntax styles that can be shared by all languages

LIME = [80, 161, 79]
BLUE = [64, 120, 242]
TURQUOISE = [1, 132, 187]
PINK = [166, 38, 164]
PINK_2 = [150, 85, 140]
YELLOW = [193, 132, 1]
SAND_YELLOW = [152, 104, 1]
RED = [244, 34, 47]
LIGHT_GREY = [160, 161, 167]
GREY = [150, 150, 150]
DARK_GREY = [56, 58, 66]

STYLES = {
    'keyword': format(PINK),
    'literals': format(TURQUOISE),
    'operator': format(GREY),
    'brace': format(DARK_GREY),
    'deffunc': format(BLUE),
    'defclass': format(YELLOW),
    'string': format(LIME),
    'string2': format(LIME),
    'meta': format(BLUE),
    'special_attributes': format(RED),
    'comment': format(LIGHT_GREY, 'italic'),
    'self': format(PINK_2, 'italic'),
    'numbers': format(SAND_YELLOW),
}


class PythonHighlighter(QSyntaxHighlighter):
    """Syntax highlighter for the Python language."""
    # Python keywords
    keywords = [
        'and', 'assert', 'break', 'class', 'continue', 'def',
        'del', 'elif', 'else', 'except', 'exec', 'finally',
        'for', 'from', 'global', 'if', 'import', 'in',
        'is', 'lambda', 'not', 'or', 'pass', 'print',
        'raise', 'return', 'try', 'while', 'yield'
    ]

    literals = [
        'False', 'True', 'None'
    ]

    # Python operators
    operators = [
        '=',
        # Comparison
        '==', '!=', '<', '<=', '>', '>=',
        # Arithmetic
        '\+', '-', '\*', '/', '//', '\%', '\*\*',
        # In-place
        '\+=', '-=', '\*=', '/=', '\%=',
        # Bitwise
        '\^', '\|', '\&', '\~', '>>', '<<',
    ]

    special_attributes = [
        '__doc__', '__name__', '__qualname__',
        '__module__', '__defaults__', '__code__',
        '__globals__', '__dict__', '__closure__',
        '__annotations__', '__kwdefaults__'
    ]

    class_dir = [
        '__class__', '__delattr__', '__dict__',
        '__dir__', '__doc__', '__eq__',
        '__format__', '__ge__', '__getattribute__',
        '__gt__', '__hash__', '__init__',
        '__init_subclass__', '__le__', '__lt__',
        '__module__', '__ne__', '__new__',
        '__reduce__', '__reduce_ex__', '__repr__',
        '__setattr__', '__sizeof__', '__str__',
        '__subclasshook__', '__weakref__'
    ]

    other = [
        'super', 'append', 'extend',
        'insert', 'remove', 'pop',
        'clear', 'index', 'count',
        'sort', 'reverse'
    ]

    types = [
        'bool', 'int', 'float',
        'complex', 'list', 'tuple',
        'range', 'str', 'bytes',
        'bytearray', 'memoryview',
        'set', 'frozenset', 'dict'
    ]

    # Python braces
    braces = [
        '\{', '\}', '\(', '\)', '\[', '\]',
    ]

    def __init__(self, document):
        QSyntaxHighlighter.__init__(self, document)

        # Multi-line strings (expression, flag, style)
        # FIXME: The triple-quotes in these two lines will mess up the
        # syntax highlighting from this point onward
        self.tri_single = (QRegExp("'''"), 1, STYLES['string2'])
        self.tri_double = (QRegExp('"""'), 2, STYLES['string2'])

        rules = []

        # Keyword, operator, and brace rules
        rules += [(r'\b%s\b' % w, 0, STYLES['keyword'])
                  for w in PythonHighlighter.keywords]
        rules += [(r'\b%s\b' % l, 0, STYLES['literals'])
                  for l in PythonHighlighter.literals]
        rules += [(r'\b%s\b' % h, 0, STYLES['meta'])
                  for h in PythonHighlighter.other]
        rules += [(r'\b%s\b' % t, 0, STYLES['meta'])
                  for t in PythonHighlighter.types]

        rules += [(r'%s' % o, 0, STYLES['operator'])
                  for o in PythonHighlighter.operators]
        rules += [(r'%s' % s, 0, STYLES['special_attributes'])
                  for s in PythonHighlighter.special_attributes]
        rules += [(r'%s' % c, 0, STYLES['meta'])
                  for c in PythonHighlighter.class_dir]
        rules += [(r'%s' % b, 0, STYLES['brace'])
                  for b in PythonHighlighter.braces]

        # All other rules
        rules += [
            # 'self'
            (r'\bself\b', 0, STYLES['self']),

            # Double-quoted string, possibly containing escape sequences
            (r'"[^"\\]*(\\.[^"\\]*)*"', 0, STYLES['string']),
            # Single-quoted string, possibly containing escape sequences
            (r"'[^'\\]*(\\.[^'\\]*)*'", 0, STYLES['string']),

            # 'def' followed by an identifier
            (r'\bdef\b\s*(\w+)', 1, STYLES['deffunc']),
            # 'class' followed by an identifier
            (r'\bclass\b\s*(\w+)', 1, STYLES['defclass']),

            # From '#' until a newline
            (r'#[^\n]*', 0, STYLES['comment']),

            # From '@' until a newline
            (r'@[^\n]*', 0, STYLES['meta']),

            # '>>>'
            (r'>>>', 0, STYLES['meta']),

            # Numeric literals
            (r'\b[+-]?[0-9]+[lL]?\b', 0, STYLES['numbers']),
            (r'\b[+-]?0[xX][0-9A-Fa-f]+[lL]?\b', 0, STYLES['numbers']),
            (r'\b[+-]?[0-9]+(?:\.[0-9]+)?(?:[eE][+-]?[0-9]+)?\b', 0, STYLES['numbers']),
        ]

        # Build a QRegExp for each pattern
        self.rules = [(QRegExp(pat), index, fmt)
                      for (pat, index, fmt) in rules]

    def highlightBlock(self, text):
        """Apply syntax highlighting to the given block of text.
        """
        # Do other syntax formatting
        for expression, nth, format in self.rules:
            index = expression.indexIn(text, 0)

            while index >= 0:
                # We actually want the index of the nth match
                index = expression.pos(nth)
                length = len(expression.cap(nth))
                self.setFormat(index, length, format)
                index = expression.indexIn(text, index + length)

        self.setCurrentBlockState(0)

        # Do multi-line strings
        in_multiline = self.match_multiline(text, *self.tri_single)
        if not in_multiline:
            in_multiline = self.match_multiline(text, *self.tri_double)

    def match_multiline(self, text, delimiter, in_state, style):
        """Do highlighting of multi-line strings. ``delimiter`` should be a
        ``QRegExp`` for triple-single-quotes or triple-double-quotes, and
        ``in_state`` should be a unique integer to represent the corresponding
        state changes when inside those strings. Returns True if we're still
        inside a multi-line string when this function is finished.
        """
        # If inside triple-single quotes, start at 0
        if self.previousBlockState() == in_state:
            start = 0
            add = 0
        # Otherwise, look for the delimiter on this line
        else:
            start = delimiter.indexIn(text)
            # Move past this match
            add = delimiter.matchedLength()

        # As long as there's a delimiter match on this line...
        while start >= 0:
            # Look for the ending delimiter
            end = delimiter.indexIn(text, start + add)
            # Ending delimiter on this line?
            if end >= add:
                length = end - start + add + delimiter.matchedLength()
                self.setCurrentBlockState(0)
            # No; multi-line string
            else:
                self.setCurrentBlockState(in_state)
                length = len(text) - start + add
            # Apply formatting
            self.setFormat(start, length, style)
            # Look for the next match
            start = delimiter.indexIn(text, start + length)

        # Return True if still inside a multi-line string, False otherwise
        if self.currentBlockState() == in_state:
            return True
        else:
            return False
